# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2023-08-26 22:39:03
# @Last Modified by:   dongqing
# @Last Modified time: 2024-04-16 13:57:19

import os
import matplotlib
import pandas as pd
import numpy as np
import scanpy as sc
import anndata as ad
import matplotlib.pyplot as plt

from scipy.spatial import KDTree
from scipy import sparse

def sub_mat(mat, genes, cells, genes_sub = None, cells_sub = None):
	mat_use = mat
	if not genes_sub:
		if type(genes[0]) == bytes:
			genes = [i.decode() for i in genes]
		genes_sub = genes	
	if not cells_sub:
		if type(genes[0]) == bytes:
			cells = [i.decode() for i in cells]
		cells_sub = cells
	genes_used = sorted(genes_sub)
	genes_used_array = np.array(genes_used)
	genes_array = np.array(genes)
	genes_sorter = np.argsort(genes_array)
	genes_used_index = genes_sorter[np.searchsorted(genes_array, genes_used_array, sorter = genes_sorter)]
	mat_use = mat_use[genes_used_index, :]
	cells_used = sorted(cells_sub)
	cells_used_array = np.array(cells_used)
	cells_array = np.array(cells)
	cells_sorter = np.argsort(cells_array)
	cells_used_index = cells_sorter[np.searchsorted(cells_array, cells_used_array, sorter = cells_sorter)]
	mat_use = mat_use[:, cells_used_index]
	return(mat_use, genes_used, cells_used)

def longdf_to_mat(gene_cell_df):
	cells = set(str(x[0]) for x in gene_cell_df.index)
	genes = set(str(x[1]) for x in gene_cell_df.index)
	cells_dict = dict(zip(cells, range(0, len(cells))))
	genes_dict = dict(zip(genes, range(0, len(genes))))
	rows = [cells_dict[str(x[0])] for x in gene_cell_df.index]
	cols = [genes_dict[str(x[1])] for x in gene_cell_df.index]
	expr_mat = sparse.csr_matrix((gene_cell_df.values, (rows, cols)))
	expr_mat= expr_mat.T
	gene_list = list(genes)
	cell_list = list(cells)
	return(expr_mat, gene_list, cell_list)

def SpotBinGene(gem_df, bin_size, countname = 'MIDCount'):
	gem_df['x_bin'] = (gem_df['x']/bin_size).astype(np.uint32)*bin_size
	gem_df['y_bin'] = (gem_df['y']/bin_size).astype(np.uint32)*bin_size
	gem_df['binID'] = gem_df['x_bin'].astype(str) + "_" + gem_df['y_bin'].astype(str)
	gem_df_bin_gene = gem_df[countname].groupby([gem_df['binID'], gem_df['geneID']]).sum()
	expr_mat, gene_list, cell_list = longdf_to_mat(gem_df_bin_gene)
	return(gem_df, expr_mat, gene_list, cell_list)

def cal_dist(coord_df, mdist):
	kd_tree = KDTree(coord_df)
	sdm = kd_tree.sparse_distance_matrix(kd_tree, mdist)
	sdm_csc = sparse.csc_matrix(sdm, dtype=np.float32)
	return(kd_tree, sdm_csc)

def cal_weight(sdm_csc):
	weight = sdm_csc.copy()
	np.reciprocal(weight.data, out=weight.data)
	np.power(weight.data, 2, out=weight.data)
	weight.setdiag(1)
	return(weight)

def enhance_spot_expr(coord_df, coord_expr_mat, neigh_dist = 5, x_colname = "x", y_colname = "y"):
	kd_tree, sdm_csc = cal_dist(coord_df[[x_colname, y_colname]], neigh_dist)
	weight = cal_weight(sdm_csc)
	coord_expr_mat_enhanced = coord_expr_mat.dot(weight)
	coord_expr_mat_enhanced_array = coord_expr_mat_enhanced.toarray()
	return(coord_expr_mat_enhanced_array)

def pearson_corr(x_array, y_array):
	xmean = x_array.mean(axis = 0)
	ymean = y_array.mean(axis = 0)
	xm = x_array - xmean
	ym = y_array - ymean
	normxm = np.linalg.norm(xm, axis = 0)
	normym = np.linalg.norm(ym, axis = 0)
	r = np.dot((xm/normxm).T, ym/normym)
	return(r)

def pearson_corr_sparse(x_array, y_array):
	xmean = x_array.mean(axis = 0)
	ymean = y_array.mean(axis = 0)
	xm = x_array - xmean
	ym = y_array - ymean
	xm = sparse.csc_matrix(xm)
	ym = sparse.csc_matrix(ym)
	normxm = sparse.linalg.norm(xm, axis = 0)
	normym = sparse.linalg.norm(ym, axis = 0)
	r = np.dot((xm/normxm).T, ym/normym)
	return(r)

def slice_coord(coord_df, span_length, overlap_length, x_colname = "x", y_colname = "y"):
	xmin = coord_df[x_colname].min()
	xmax = coord_df[x_colname].max()
	ymin = coord_df[y_colname].min()
	ymax = coord_df[y_colname].max()
	i = xmin
	coord_sliced_list = []
	while i + span_length <= xmax:
		xmin_sliced = i
		xmax_sliced = i + span_length
		j = ymin
		while j + span_length <= ymax:
			ymin_sliced = j
			ymax_sliced = j + span_length
			coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
			j = j + span_length - overlap_length
		ymin_sliced = j
		ymax_sliced = ymax
		coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
		i = i + span_length - overlap_length
	xmin_sliced = i
	xmax_sliced = xmax + 1
	j = ymin
	while j + span_length <= ymax:
		ymin_sliced = j
		ymax_sliced = j + span_length
		coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
		j = j + span_length - overlap_length
	ymin_sliced = j
	ymax_sliced = ymax + 1
	coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
	return(coord_sliced_list)

def filter_mat(all_count_mat, all_count_genes, all_count_spots, sub_spots = None, gene_pct = 0.05):
	if sub_spots:
		sub_expr, sub_gene, sub_spot = sub_mat(mat = all_count_mat, 
											   genes = all_count_genes, 
											   cells = all_count_spots, 
											   cells_sub = sub_spots)
	else:
		sub_expr, sub_gene, sub_spot = all_count_mat, all_count_genes, all_count_spots
	sub_gene_inspot = sub_expr.sum(axis = 1)
	nSpot = np.quantile(np.array(sub_gene_inspot.flatten())[0], (1 - gene_pct))
	sub_gene_over = np.array(sub_gene)[np.array((sub_gene_inspot > nSpot).flatten())[0]]
	sub_expr_over, sub_gene_over_list, sub_spot_over = sub_mat(mat = sub_expr, 
															   genes = sub_gene, 
															   cells = sub_spot, 
															   genes_sub = sub_gene_over.tolist())
	sub_spot_ngene = sub_expr.sum(axis = 0)
	sub_spot_over0 = np.array(sub_spot_over)[np.array((sub_spot_ngene > 0).flatten())[0]]
	sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list = sub_mat(mat = sub_expr_over, 
															   genes = sub_gene_over_list, 
															   cells = sub_spot_over, 
															   cells_sub = sub_spot_over0.tolist())
	return(sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list)

def get_hvg(all_count_mat, all_count_genes, all_count_spots, n_top_genes):
	all_count_ad = ad.AnnData(
		all_count_mat.T,
		obs=pd.DataFrame(index=all_count_spots),
		var=pd.DataFrame(index=all_count_genes),
	)
	# sc.pp.filter_cells(all_count_ad, min_genes=200)
	# sc.pp.filter_genes(all_count_ad, min_cells=3)
	# sc.pp.normalize_total(all_count_ad, target_sum=1e4)
	# sc.pp.log1p(all_count_ad)
	sc.pp.highly_variable_genes(all_count_ad, n_top_genes=n_top_genes, flavor='seurat_v3', span=1.0)
	hvg = all_count_ad.var[all_count_ad.var.highly_variable].index.tolist()
	return(hvg)

def get_frequent_gene(all_count_mat, all_count_genes, all_count_spots, gene_pct):
	all_count_mat[all_count_mat > 0] = 1
	gene_inspot = all_count_mat.sum(axis = 1)
	nSpot = np.quantile(np.array(gene_inspot.flatten())[0], (1 - gene_pct))
	freq_gene = np.array(all_count_genes)[np.array((gene_inspot > nSpot).flatten())[0]]
	return(freq_gene.tolist())

def get_cell_dist(props_df_sub):
	cell_cell_dist = cal_dist(props_df_sub.loc[:,('centroid-0','centroid-1')], 100)[1]
	cell_cell_dist_array = cell_cell_dist.toarray()
	props_df_sub.loc[:, 'min_cell_dist'] = np.amin(cell_cell_dist_array, axis = 1, where = cell_cell_dist_array!=0, initial = 100)
	props_df_sub.loc[:, 'min_cell_dist'] = np.round(props_df_sub['min_cell_dist'], 2)
	cell_cell_dist_median = np.median(props_df_sub['min_cell_dist'])
	max_dist = cell_cell_dist_median/2
	return(max_dist)

def filter_mat_by_var(all_count_mat, all_count_genes, all_count_spots, sub_spots = None):
	if sub_spots:
		sub_expr, sub_gene, sub_spot = sub_mat(mat = all_count_mat, 
											   genes = all_count_genes, 
											   cells = all_count_spots, 
											   cells_sub = sub_spots)
	else:
		sub_expr, sub_gene, sub_spot = all_count_mat, all_count_genes, all_count_spots
	all_count_ad = ad.AnnData(
		all_count_mat.T,
		obs=pd.DataFrame(index=all_count_spots),
		var=pd.DataFrame(index=all_count_genes),
	)
	sc.pp.highly_variable_genes(all_count_ad, n_top_genes=2000, flavor='seurat_v3', span=1.0)
	selected_gene = all_count_ad.var[all_count_ad.var.highly_variable].index
	sub_expr_over, sub_gene_over_list, sub_spot_over = sub_mat(mat = sub_expr, 
															   genes = sub_gene, 
															   cells = sub_spot, 
															   genes_sub = selected_gene.tolist())
	sub_spot_ngene = sub_expr_over.sum(axis = 0)
	sub_spot_over0 = np.array(sub_spot_over)[np.array((sub_spot_ngene > 0).flatten())[0]]
	sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list = sub_mat(mat = sub_expr_over, 
															   genes = sub_gene_over_list, 
															   cells = sub_spot_over, 
															   cells_sub = sub_spot_over0.tolist())
	return(sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list)

def cal_corr_within_seg(count_df_sub, seg_res, count_name = 'MIDCount', genes_list = None):
	count_df_sub_in = count_df_sub[count_df_sub[seg_res].notna()]
	count_df_sub_in_1 = count_df_sub[count_df_sub['Random_divide'] == 1]
	count_df_sub_in_2 = count_df_sub[count_df_sub['Random_divide'] == 2]
	gene_seg_cell_1 = count_df_sub_in_1[count_name].groupby([count_df_sub_in_1[seg_res], count_df_sub_in_1['geneID']]).sum()
	seg_count_mat_1, seg_count_genes_1, seg_count_cells_1 = longdf_to_mat(gene_seg_cell_1)
	gene_seg_cell_2 = count_df_sub_in_2[count_name].groupby([count_df_sub_in_2[seg_res], count_df_sub_in_2['geneID']]).sum()
	seg_count_mat_2, seg_count_genes_2, seg_count_cells_2 = longdf_to_mat(gene_seg_cell_2)
	if genes_list:
		genes_overlap = list((set(genes_list) & set(seg_count_genes_1)) & set(seg_count_genes_2))
	else:
		genes_overlap = list(set(seg_count_genes_1) & set(seg_count_genes_2))
	cells_overlap = list(set(seg_count_cells_1) & set(seg_count_cells_2))
	seg_count_mat_1_overlap, seg_count_gene_1_overlap, seg_count_cell_1_overlap = sub_mat(mat = seg_count_mat_1, 
								   genes = seg_count_genes_1, 
								   cells = seg_count_cells_1, 
								   cells_sub = cells_overlap, 
								   genes_sub = genes_overlap)
	seg_count_mat_2_overlap, seg_count_gene_2_overlap, seg_count_cell_2_overlap = sub_mat(mat = seg_count_mat_2, 
								   genes = seg_count_genes_2, 
								   cells = seg_count_cells_2, 
								   cells_sub = cells_overlap, 
								   genes_sub = genes_overlap)
	corr_random = pearson_corr(seg_count_mat_1_overlap.toarray(), seg_count_mat_2_overlap.toarray()).diagonal()
	return(corr_random, seg_count_cell_1_overlap)

def cal_corr_within_seg_nucleus(count_df_sub, seg_res, genes_list, count_name = 'MIDCount'):
	count_df_sub_in = count_df_sub[count_df_sub[seg_res].notna()]
	count_df_sub_in_1 = count_df_sub_in[count_df_sub_in['Nucleus'] == 0]
	count_df_sub_in_2 = count_df_sub_in[count_df_sub_in['Nucleus'] == 1]
	gene_seg_cell_1 = count_df_sub_in_1[count_name].groupby([count_df_sub_in_1[seg_res], count_df_sub_in_1['geneID']]).sum()
	seg_count_mat_1, seg_count_genes_1, seg_count_cells_1 = longdf_to_mat(gene_seg_cell_1)
	gene_seg_cell_2 = count_df_sub_in_2[count_name].groupby([count_df_sub_in_2[seg_res], count_df_sub_in_2['geneID']]).sum()
	seg_count_mat_2, seg_count_genes_2, seg_count_cells_2 = longdf_to_mat(gene_seg_cell_2)
	genes_overlap = list((set(genes_list) & set(seg_count_genes_1)) & set(seg_count_genes_2))
	cells_overlap = list(set(seg_count_cells_1) & set(seg_count_cells_2))
	seg_count_mat_1_overlap, seg_count_gene_1_overlap, seg_count_cell_1_overlap = sub_mat(mat = seg_count_mat_1, 
	                               genes = seg_count_genes_1, 
	                               cells = seg_count_cells_1, 
	                               cells_sub = cells_overlap, 
	                               genes_sub = genes_overlap)
	seg_count_mat_2_overlap, seg_count_gene_2_overlap, seg_count_cell_2_overlap = sub_mat(mat = seg_count_mat_2, 
	                               genes = seg_count_genes_2, 
	                               cells = seg_count_cells_2, 
	                               cells_sub = cells_overlap, 
	                               genes_sub = genes_overlap)
	corr_random = pearson_corr(seg_count_mat_1_overlap.toarray(), seg_count_mat_2_overlap.toarray()).diagonal()
	return((corr_random, seg_count_cell_2_overlap))

def cal_corr_between_seg(count_df_sub, seg_res_1, seg_res_2, genes_list = None):
	count_df_sub_overlap = count_df_sub.loc[count_df_sub[seg_res_1] == count_df_sub[seg_res_2], :]
	count_df_sub_nonoverlap = count_df_sub.loc[count_df_sub[seg_res_1] != count_df_sub[seg_res_2], :]
	# overlap
	count_df_sub_overlap = count_df_sub_overlap[count_df_sub_overlap[seg_res_1].notna()]
	gene_seg_cell_overlap = count_df_sub_overlap['MIDCount'].groupby([count_df_sub_overlap[seg_res_1], count_df_sub_overlap['geneID']]).sum()
	seg_count_mat_overlap, seg_count_genes_overlap, seg_count_cells_overlap = longdf_to_mat(gene_seg_cell_overlap)
	# segmentation 1
	count_df_sub_nonoverlap_1 = count_df_sub_nonoverlap[count_df_sub_nonoverlap[seg_res_1].notna()]
	gene_seg_cell_1 = count_df_sub_nonoverlap_1['MIDCount'].groupby([count_df_sub_nonoverlap_1[seg_res_1], count_df_sub_nonoverlap_1['geneID']]).sum()
	seg_count_mat_1, seg_count_genes_1, seg_count_cells_1 = longdf_to_mat(gene_seg_cell_1)
	# segmentation 2
	count_df_sub_nonoverlap_2 = count_df_sub_nonoverlap[count_df_sub_nonoverlap[seg_res_2].notna()]
	gene_seg_cell_2 = count_df_sub_nonoverlap_2['MIDCount'].groupby([count_df_sub_nonoverlap_2[seg_res_2], count_df_sub_nonoverlap_2['geneID']]).sum()
	seg_count_mat_2, seg_count_genes_2, seg_count_cells_2 = longdf_to_mat(gene_seg_cell_2)
	# determine cells and genes to use
	cells_overlap = sorted(np.intersect1d(np.intersect1d(seg_count_cells_overlap, seg_count_cells_1), seg_count_cells_2).tolist())
	genes_overlap = sorted(np.intersect1d(np.intersect1d(seg_count_genes_overlap, seg_count_genes_1), seg_count_genes_2).tolist())
	if genes_list:
		genes_use = sorted(np.intersect1d(genes_overlap, np.array(genes_list)).tolist())
	else:
		genes_use = genes_overlap
	seg_count_mat_overlap_use, seg_count_genes_overlap_use, seg_count_cells_overlap_use = sub_mat(
											  mat = seg_count_mat_overlap, 
											  genes = seg_count_genes_overlap, 
											  cells = seg_count_cells_overlap, 
											  cells_sub = cells_overlap, 
											  genes_sub = genes_use)
	seg_count_mat_1_use, seg_count_genes_1_use, seg_count_cells_1_use = sub_mat(
											  mat = seg_count_mat_1, 
											  genes = seg_count_genes_1, 
											  cells = seg_count_cells_1, 
											  cells_sub = cells_overlap, 
											  genes_sub = genes_use)
	seg_count_mat_2_use, seg_count_genes_2_use, seg_count_cells_2_use = sub_mat(
											  mat = seg_count_mat_2, 
											  genes = seg_count_genes_2, 
											  cells = seg_count_cells_2, 
											  cells_sub = cells_overlap, 
											  genes_sub = genes_use)
	corr_1 = pearson_corr(seg_count_mat_overlap_use.toarray(), seg_count_mat_1_use.toarray()).diagonal()
	corr_2 = pearson_corr(seg_count_mat_overlap_use.toarray(), seg_count_mat_2_use.toarray()).diagonal()
	return((corr_1, corr_2, seg_count_cells_2_use))

def KL_divergence(X, Y):
	X = X + 0.01
	Y = Y + 0.01
	X = X/X.sum(axis=1, keepdims=True)
	Y = Y/Y.sum(axis=1, keepdims=True)
	log_X = np.log(X)
	log_Y = np.log(Y)
	X_log_X = np.matrix((X*log_X).sum(axis = 1))
	D = X_log_X.T - np.dot(X,log_Y.T)
	return np.asarray(D)

def KL_divergence_rm0(X, Y):
	d_list = []
	for x_idx in range(X.shape[0]):
		d_x_list = []
		for y_idx in range(Y.shape[0]):
			x = X[x_idx, :]
			y = Y[y_idx, :]
			gene_union = np.union1d(np.where(x !=0), np.where(y !=0))
			x = x[gene_union]
			y = y[gene_union]
			x = x + 0.01
			y = y + 0.01
			x = x/x.sum()
			y = y/y.sum()
			log_x = np.log(x)
			log_y = np.log(y)
			x_log_x = (x*log_x).sum()
			d = x_log_x.T - np.dot(x,log_y.T)
			d_x_list.append(d)
		d_list.append(d_x_list)
	D = np.array(d_list)
	return(D)

def get_cell_mat(count_df_seg, seg_res, count_name = "MIDCount"):
	count_df_seg = count_df_seg.dropna(subset = [seg_res])
	gene_cell = count_df_seg[count_name].groupby([count_df_seg[seg_res], count_df_seg['geneID']]).sum()
	expr_mat, gene_list, cell_list = longdf_to_mat(gene_cell)
	return(expr_mat, gene_list, cell_list)



# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2023-10-14 12:05:31
# @Last Modified by:   dongqing
# @Last Modified time: 2024-04-15 19:21:03

import os
import h5py
import tables
import collections
import numpy as np
import pandas as pd
from scipy import sparse
from Cellist.Utility import get_cell_mat, longdf_to_mat

FeatureBCMatrix = collections.namedtuple('FeatureBCMatrix', ['ids', 'names', 'barcodes', 'matrix'])

def read_10X_h5(filename):
    """Read 10X HDF5 files, support both gene expression and peaks."""
    with tables.open_file(filename, 'r') as f:
        try:
            group = f.get_node(f.root, 'matrix')
        except tables.NoSuchNodeError:
            print("Matrix group does not exist in this file.")
            return None
        feature_group = getattr(group, 'features')
        ids = getattr(feature_group, 'id').read()
        names = getattr(feature_group, 'name').read()
        barcodes = getattr(group, 'barcodes').read()
        data = getattr(group, 'data').read()
        indices = getattr(group, 'indices').read()
        indptr = getattr(group, 'indptr').read()
        shape = getattr(group, 'shape').read()
        matrix = sparse.csc_matrix((data, indices, indptr), shape=shape)
        return FeatureBCMatrix(ids, names, barcodes, matrix)

def write_10X_h5(filename, matrix, features, barcodes, datatype = 'Peak'):
    """Write 10X HDF5 files, support both gene expression and peaks."""
    f = h5py.File(filename, 'w')
    if datatype == 'Peak':
       M = sparse.csc_matrix(matrix, dtype=np.int8)
    else:
       M = sparse.csc_matrix(matrix, dtype=np.float32)
    B = np.array(barcodes, dtype='|S200')
    P = np.array(features, dtype='|S100')
    FT = np.array([datatype]*len(features), dtype='|S100')
    AT = np.array(['genome'], dtype='|S10')
    mat = f.create_group('matrix')
    mat.create_dataset('barcodes', data=B)
    mat.create_dataset('data', data=M.data)
    mat.create_dataset('indices', data=M.indices)
    mat.create_dataset('indptr', data=M.indptr)
    mat.create_dataset('shape', data=M.shape)
    fet = mat.create_group('features')
    fet.create_dataset('_all_tag_keys', data=AT)
    fet.create_dataset('feature_type', data=FT)
    fet.create_dataset('id', data=P)
    fet.create_dataset('name', data=P)
    f.close()

def read_h5(h5_file):
    expr_read = read_10X_h5(filename = h5_file)
    expr_mat = expr_read.matrix
    expr_genes = expr_read.names.tolist()
    expr_cells = expr_read.barcodes.tolist()
    if type(expr_genes[0]) == bytes:
        expr_genes = [i.decode() for i in expr_genes]
    if type(expr_cells[0]) == bytes:
        expr_cells = [i.decode() for i in expr_cells]
    return(expr_mat, expr_genes, expr_cells)

def write_segmentation_h5(count_df_seg, seg_res, out_prefix, out_dir, count_name):
    expr_mat, gene_list, cell_list = get_cell_mat(count_df_seg, seg_res, count_name)
    count_h5_file = os.path.join(out_dir, "%s_segmentation_cell_count.h5" %out_prefix)
    write_10X_h5(filename = count_h5_file, matrix = expr_mat, features = gene_list, barcodes = cell_list, datatype = 'Gene')

def write_segmentation_cell_coord(coord_df_seg, seg_res, out_prefix, out_dir):
    coord_df_seg = coord_df_seg.dropna(subset = [seg_res])
    cell_coord = coord_df_seg[['x', 'y', seg_res]].groupby(seg_res).mean()
    cell_nspot = coord_df_seg[['x',seg_res]].groupby(seg_res).count()
    cell_nspot = cell_nspot.rename(columns = {'x': "nSpot"})
    cell_coord_nspot = pd.merge(cell_coord, cell_nspot, on = seg_res)
    cell_coord_file = os.path.join(out_dir, "%s_segmentation_cell_coord.txt" %out_prefix)
    cell_coord_nspot.to_csv(cell_coord_file, sep = "\t")

def gem_to_mat(gem_df, outfile, countname = 'MIDCount'):
    gem_df['x_y'] = gem_df['x'].astype(str) + '_' + gem_df['y'].astype(str)
    gene_spot = gem_df[countname].groupby([gem_df['x_y'], gem_df['geneID']]).sum()
    spot_expr_mat, gene_list, spot_list = longdf_to_mat(gene_spot)
    count_h5_file = outfile
    write_10X_h5(filename = count_h5_file, matrix = spot_expr_mat, features = gene_list, barcodes = spot_list,  datatype = 'Gene')
    return(gem_df)



# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2024-01-05 20:29:00
# @Last Modified by:   dongqing
# @Last Modified time: 2024-04-17 12:28:57


import sys, os, gc
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

import matplotlib
matplotlib.rcParams['font.family'] = 'sans-serif'
matplotlib.rcParams['font.sans-serif'] = ['Arial']

import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s  %(message)s')

from skimage.io import imread, imsave
from skimage.filters import threshold_otsu, threshold_multiotsu, threshold_niblack, threshold_local
from skimage.feature import peak_local_max
from skimage.measure import label, regionprops_table
from skimage.segmentation import watershed, expand_labels, mark_boundaries
from skimage.color import label2rgb, rgb2gray, rgb2hed
from skimage.segmentation import mark_boundaries
from skimage.morphology import remove_small_holes

from scipy.sparse import csc_matrix, save_npz, load_npz, csr_matrix
from scipy import ndimage as ndi

from Cellist.Utility import *
from Cellist.IO import write_10X_h5, write_segmentation_h5, write_segmentation_cell_coord, gem_to_mat
from Cellist.Plot import *

'''
cellist watershed --gem /mnt/Venus/home/dongxin/Projects/heart/202401_Stereo_seq/result_C02942B3/041.cellcut/C02942B3.adjusted.gem \
--tif Alignment/ssDNA_C02942B3_regist_regist_transposed_aligned_by_Spateo.tiff \
--outdir Watershed --outprefix C02942B3

cellist watershed --gem /fs/home/dongxin/Projects/heart/202401_Stereo_seq/result_C02942B3/041.cellcut/C02942B3.adjusted.gem \
--tif Alignment/ssDNA_C02942B3_regist_regist_transposed_aligned_by_Spateo.tiff \
--outdir Watershed --outprefix C02942B3
'''

def WatershedParser(subparsers):
    parser = subparsers.add_parser("watershed", 
        help = "Run initial watershed segmentation on the staining image. ")
    group_input = parser.add_argument_group("Input arguments")
    group_input.add_argument("--platform", dest = "platform", default = 'barcoding', choices = ['barcoding', 'imaging'],
        help = "Supported high-resoultion ST platform: 'barcoding' (e.g., Stereo-seq, Seq-scope) or "
        "'imaging' (e.g., seqFISH, STARmap, 10x Xenium). Default: barcoding.")
    group_input.add_argument("--gem", dest = "spot_expr_file", default = None,
        help = "Bin1 gem file generated by SAW pipeline located in '041.cellcut' or '04.tissuecut'. ")
    group_input.add_argument("--tif", dest = "regist_tif_file", default = None,
        help = "Registered tif file generated by SAW pipeline located in '04.tissuecut' or the refined tif file generated by 'cellist align'. ")
    group_input.add_argument("--min-distance", dest = "min_distance", type = int, default = 6,
        help = "The minimal allowed distance (in pixel) separating peaks for watershed segmentation. "
        "Peaks are the local maxima in a region of 2 * min_distance + 1 (i.e. peaks are separated by at least min_distance). Default: 6. ")
    group_input.add_argument("--no-local-threshold", dest = "no_local_threshold", action = "store_true", 
        help = "Whether or not to perform local thresholding on the image. "
        "If not set, Cellist will perform local thresholding. If set, the local thresholding will be skipped. ")

    group_output = parser.add_argument_group("Output arguments")
    group_output.add_argument("--outdir", dest = "out_dir", default = '.',
        help = "The output directory. ")
    group_output.add_argument("--outprefix", dest = "out_prefix", default = None,
        help = "The output prefix. ")
    return(parser)


def Thresholding(platform, img_path, out_dir, out_img_prefix, no_local_threshold = False):
    img = imread(img_path)
    logging.info("Read image done. ")
    # 0. multiotsu thresholding
    logging.info("Thresholding... ")
    threshold = threshold_multiotsu(img, 3)[0]
    mask = img > threshold
    img_mask = img.copy()
    img_mask[~mask] = 0
    if no_local_threshold: 
        otsu_local = img_mask.copy()
        otsu_local_hole_rm = remove_small_holes(otsu_local)
        fig, ax = plt.subplots(1, 2, figsize=(30, 17), sharex=True, sharey=True)
        ax[0].imshow(img, cmap="gray")
        ax[0].set_title("Original", fontsize = 20)
        ax[1].imshow(mask, cmap = "gray")
        ax[1].set_title('Mask', fontsize = 20)
        for a in ax.ravel():
            a.axis('off')
        plt.savefig(os.path.join(out_dir, "%s_thresholding.png" %out_img_prefix), bbox_inches='tight')
    else:
        # 1. local thresholding
        local_thresh = threshold_local(img, block_size = 51, offset=0)
        local = img > local_thresh
        # 2. combine two thresholding results
        mask_local = img_mask > local_thresh
        otsu_local = np.logical_and(local, mask_local)        
        # 3. remove small holes
        otsu_local_hole_rm = remove_small_holes(otsu_local)
        if platform == "imaging":
            otsu_local_hole_rm = mask.copy()
        fig, ax = plt.subplots(2, 2, figsize=(30, 34), sharex=True, sharey=True)
        ax[0, 0].imshow(img, cmap="gray")
        ax[0, 0].set_title("Original", fontsize = 20)
        ax[0, 1].imshow(mask, cmap = "gray")
        ax[0, 1].set_title('Mask', fontsize = 20)
        ax[1, 0].imshow(local, cmap = "gray")
        ax[1, 0].set_title('Local thresholding', fontsize = 20)
        ax[1, 1].imshow(otsu_local_hole_rm, cmap = "gray")
        ax[1, 1].set_title('Mask + local thresholding', fontsize = 20)
        for a in ax.ravel():
            a.axis('off')
        plt.savefig(os.path.join(out_dir, "%s_thresholding.png" %out_img_prefix), bbox_inches='tight')
    imsave(os.path.join(out_dir, "%s_thresholding.tif" %out_img_prefix), otsu_local_hole_rm)
    logging.info("Threshold done. ")
    return img, otsu_local_hole_rm

def watershed_seg(img, otsu_local_hole_rm, out_dir, out_img_prefix, min_distance = 6, expansion = False, expansion_dist = 8):
    logging.info("Wartershed segmentation... ")
    # 0. find local markers
    distance_local = ndi.distance_transform_edt(otsu_local_hole_rm)
    local_max_coords = peak_local_max(distance_local, min_distance = min_distance, exclude_border=False)
    local_max_mask = np.zeros(distance_local.shape, dtype=bool)
    local_max_mask[tuple(local_max_coords.T)] = True
    markers = label(local_max_mask)
    fig, ax = plt.subplots(figsize=(20, 20))
    ax.imshow(img, cmap='gray')
    ax.plot(local_max_coords[:,1], local_max_coords[:,0], 'r*', markersize = 2)
    ax.set_title('Markers (%s local maxima)' % local_max_coords.shape[0], fontsize = 20)
    plt.savefig(os.path.join(out_dir, "%s_markers.png" %out_img_prefix), bbox_inches='tight')
    # 1. watershed segmentation
    segmented_cells = watershed(-distance_local, markers, mask = otsu_local_hole_rm)
    logging.info("Watershed number of segments: %s" %(len(np.unique(segmented_cells))))
    watershed_seg_file = os.path.join(out_dir, "%s_watershed_nucleus_matrix.npz" %out_img_prefix)
    segmented_cells_mat = csc_matrix(segmented_cells, dtype=np.int32)
    save_npz(watershed_seg_file, segmented_cells_mat)
    fig, ax = plt.subplots(figsize=(20, 20))
    ax.imshow(mark_boundaries(img, segmented_cells, color = (1,0,0)))
    plt.savefig(os.path.join(out_dir, "%s_cell_boundary.png" %out_img_prefix), bbox_inches='tight')
    props = regionprops_table(segmented_cells, 
                              properties=['label', 'area','centroid',
                                          'equivalent_diameter_area'])
    props_df = pd.DataFrame(props)
    props_df.to_csv(os.path.join(out_dir, "%s_watershed_nucleus_property.txt" %out_img_prefix), sep = "\t", index = False)
    if expansion:
        expanded = expand_labels(segmented_cells, distance = expansion_dist)
        watershed_seg_file = os.path.join(out_dir, "%s_watershed_cells_expansion%s_matrix.npz" %(out_img_prefix, expansion_dist))
        expanded_mat = csc_matrix(expanded, dtype=np.int32)
        save_npz(watershed_seg_file, expanded_mat)
        props_expanded = regionprops_table(expanded, 
                                  properties=['label', 'area','centroid',
                                              'equivalent_diameter_area'])
        props_df_expanded = pd.DataFrame(props_expanded)
        props_df_expanded.to_csv(os.path.join(out_dir, "%s_watershed_cells_expansion%s_property.txt" %(out_img_prefix, expansion_dist)), sep = "\t", index = False)
        return segmented_cells, expanded
    else:
        return segmented_cells

def write_segmentation_coord(segmented_cells, coord_df, out_dir, out_gem_prefix, expanded = None, seg_method = "nucleus", expansion_dist = 8):
    logging.info("Writing segmentation results... ")
    segmented_nucleus_df = pd.DataFrame(segmented_cells)
    segmented_nucleus_df["x"] = segmented_nucleus_df.index
    segmented_nucleus_long_df = pd.melt(segmented_nucleus_df, id_vars = "x", 
                                         value_vars = segmented_nucleus_df.columns[0:-1],
                                         var_name = "y",
                                         value_name = "Watershed")
    del segmented_nucleus_df
    gc.collect()
    segmented_nucleus_long_df["Nucleus"] = 0
    segmented_nucleus_long_df.loc[segmented_nucleus_long_df["Watershed"] > 0, "Nucleus"] = 1
    xmin,xmax = coord_df['x'].min(), coord_df['x'].max()
    ymin,ymax = coord_df['y'].min(), coord_df['y'].max()
    segmented_nucleus_long_df = segmented_nucleus_long_df.loc[(segmented_nucleus_long_df['x'] >= xmin) & 
                                                          (segmented_nucleus_long_df['x'] <= xmax) &
                                                          (segmented_nucleus_long_df['y'] >= ymin) &
                                                          (segmented_nucleus_long_df['y'] <= ymax),:]

    x_list = segmented_nucleus_long_df.loc[:,'x'].tolist()
    y_list = segmented_nucleus_long_df.loc[:,'y'].tolist()
    x_y = [f"{a}_{b}" for a, b in zip(x_list, y_list)]
    segmented_nucleus_long_df['x_y'] = x_y
    # segmented_nucleus_long_df['x_y'] = segmented_nucleus_long_df['x'].astype(str) + '_' + segmented_nucleus_long_df['y'].astype(str)
    coord_slice_list = slice_coord(coord_df, 5000, 0)
    segmented_nucleus_long_df_barcode_sub_list = []
    for coord_tuple in coord_slice_list:
        xmin_slice, xmax_slice, ymin_slice, ymax_slice = coord_tuple
        coord_df_sub = coord_df.loc[(coord_df['x'] >= xmin_slice) & 
                      (coord_df['x'] < xmax_slice) &
                      (coord_df['y'] >= ymin_slice) &
                      (coord_df['y'] < ymax_slice),:]
        segmented_nucleus_long_df_sub = segmented_nucleus_long_df.loc[(segmented_nucleus_long_df['x'] >= xmin_slice) & 
                      (segmented_nucleus_long_df['x'] < xmax_slice) &
                      (segmented_nucleus_long_df['y'] >= ymin_slice) &
                      (segmented_nucleus_long_df['y'] < ymax_slice),:]
        segmented_nucleus_long_df_barcode_sub = pd.merge(coord_df_sub, segmented_nucleus_long_df_sub[['x_y','Nucleus','Watershed']], how = 'left', on = 'x_y')
        segmented_nucleus_long_df_barcode_sub_list.append(segmented_nucleus_long_df_barcode_sub)
    segmented_nucleus_long_df_barcode = pd.concat(segmented_nucleus_long_df_barcode_sub_list)
    # segmented_nucleus_long_df_barcode = pd.merge(coord_df, segmented_nucleus_long_df[['x_y','Nucleus','Watershed']], how = 'left', on = 'x_y')
    segmented_nucleus_long_df_barcode = segmented_nucleus_long_df_barcode.sort_values(['x', 'y'])
    nucl_coord_out_file = os.path.join(out_dir, "%s_watershed_nucleus_coord.txt" %(out_gem_prefix))
    segmented_nucleus_long_df_barcode.to_csv(nucl_coord_out_file, sep = "\t", index = False)
    logging.info("Writing segmentation results is done.")
    if expanded:
        segmented_cells_df = pd.DataFrame(expanded)
        segmented_cells_df["x"] = segmented_cells_df.index
        segmented_cells_long_df = pd.melt(segmented_cells_df, id_vars = "x", 
                                         value_vars = segmented_cells_df.columns[0:-1],
                                         var_name = "y",
                                         value_name = "Expansion")
        segmented_cells_long_df["Nucleus"] = 0
        segmented_cells_long_df.loc[segmented_nucleus_long_df["Watershed"] > 0, "Nucleus"] = 1
        segmented_cells_long_df['x_y'] = segmented_cells_long_df['x'].astype(str) + '_' + segmented_cells_long_df['y'].astype(str)
        segmented_cells_long_df_barcode = pd.merge(coord_df, segmented_cells_long_df[['x_y','Nucleus','Expansion',]], how = 'left', on = 'x_y')
        segmented_cells_long_df_barcode = segmented_cells_long_df_barcode.sort_values(['x', 'y'])
        segmented_cells_long_df_barcode.loc[segmented_cells_long_df_barcode['Expansion'] == 0, 'Expansion'] = np.nan
        cell_coord_out_file = os.path.join(out_dir, "%s_watershed_cells_expansion%s_coord.txt" %(out_gem_prefix, expansion_dist))
        segmented_cells_long_df_barcode.to_csv(cell_coord_out_file, sep = "\t", index = False)
        return segmented_nucleus_long_df_barcode, segmented_cells_long_df_barcode
    else:
        return segmented_nucleus_long_df_barcode

def write_segmentation_cell(segmented_nucleus_long_df_barcode, count_df, out_gem_prefix, out_dir, segmented_cells_long_df_barcode = None):
    logging.info("Writing cell-level expression matrix... ")
    count_df_watershed = pd.merge(count_df, segmented_nucleus_long_df_barcode[['x_y', 'Watershed']], how = "right", on = 'x_y')
    count_name = count_df_watershed.columns[3]
    count_df_watershed = count_df_watershed.loc[count_df_watershed['Watershed'] != 0, :]
    write_segmentation_h5(count_df_seg = count_df_watershed, seg_res = "Watershed", out_prefix = "%s_waterhsed" %out_gem_prefix, 
        out_dir = out_dir, count_name = count_name)
    write_segmentation_cell_coord(coord_df_seg = segmented_nucleus_long_df_barcode, seg_res = 'Watershed', 
        out_prefix = "%s_waterhsed" %out_gem_prefix, out_dir = out_dir)
    if segmented_cells_long_df_barcode:
        count_df_expansion = pd.merge(count_df, segmented_cells_long_df_barcode[['x_y', 'Expansion']], how = "right", on = 'x_y')
        count_name = count_df_expansion.columns[3]
        write_segmentation_h5(count_df_seg = count_df_watershed, seg_res = "Expansion", out_prefix = "%s_expansion" %out_gem_prefix, 
            out_dir = out_dir, count_name = count_name)
        write_segmentation_cell_coord(coord_df_seg = segmented_cells_long_df_barcode, seg_res = 'Expansion', 
            out_prefix = "%s_expansion" %out_gem_prefix, out_dir = out_dir)
    logging.info("Writing cell-level expression matrix is done.")

def Watershed(platform, gem_path, img_path, out_dir, out_prefix, min_distance, no_local_threshold):
    if not out_prefix:
        file_name, file_extension = os.path.splitext(gem_path)
        base_name = os.path.basename(file_name)
        out_gem_prefix = base_name
        file_name, file_extension = os.path.splitext(img_path)
        base_name = os.path.basename(file_name)
        out_img_prefix = base_name
    else:
        out_gem_prefix = out_prefix
        out_img_prefix = out_prefix
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    # --------------------- read image file and thresholding --------------------- 
    threshold_file = os.path.join(out_dir, "%s_thresholding.tif" %out_img_prefix)
    if os.path.exists(threshold_file):
        otsu_local_hole_rm = imread(threshold_file)
        img = imread(img_path)
    else:
        img, otsu_local_hole_rm = Thresholding(platform, img_path, out_dir, out_img_prefix, no_local_threshold)
    # --------------------- watershed segmentation --------------------- 
    segmented_cells = watershed_seg(img, otsu_local_hole_rm, out_dir, out_img_prefix, min_distance = min_distance, expansion = False, expansion_dist = 8)
    # --------------------- read gem file --------------------- 
    logging.info("Reading count file...")
    gem_df = pd.read_csv(gem_path, sep = "\t", comment = "#")
    bin1_out_file = os.path.join(out_dir, "%s_bin1.h5" %out_gem_prefix)
    countname = gem_df.columns[3]
    gem_df = gem_to_mat(gem_df = gem_df, outfile = bin1_out_file, countname = countname)
    coord_df = gem_df.drop_duplicates(subset = ['x_y'])
    coord_df = coord_df[["x", "y", "x_y"]]
    # --------------------- write segmentation result --------------------- 
    segmented_nucleus_long_df_barcode = write_segmentation_coord(segmented_cells, coord_df, out_dir, out_gem_prefix)
    write_segmentation_cell(segmented_nucleus_long_df_barcode, gem_df, out_gem_prefix, out_dir)
    logging.info("Drawing segmentation plot...")
    draw_segmentation(segmented_nucleus_long_df_barcode, seg_res = "Watershed", out_prefix = "%s_watershed" %out_gem_prefix,
        out_dir = out_dir, x = "x", y = "y", figsize = (80, 80))
    logging.info("Drawing segmentation plot is done.")
    logging.info("All done!")

if __name__ == '__main__':
    parser = WatershedParser()
    platform = parser.platform
    gem_path = parser.spot_expr_file
    img_path = parser.regist_tif_file
    out_dir = parser.out_dir
    out_prefix = parser.out_prefix
    min_distance = parser.min_distance
    no_local_threshold = parser.no_local_threshold
    Watershed(platform = platform, gem_path = gem_path, img_path = img_path, out_dir = out_dir, out_prefix = out_prefix, min_distance = min_distance, no_local_threshold = no_local_threshold)



# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2024-01-07 19:26:42
# @Last Modified by:   dongqing
# @Last Modified time: 2024-04-17 03:43:30

import os, sys
import argparse
import functools
import pandas as pd
import numpy as np
import multiprocessing as mp
import scanpy as sc
import anndata
import concurrent.futures

from sklearn_ann.kneighbors.annoy import AnnoyTransformer
from Cellist.IO import read_10X_h5, write_10X_h5
from annoy import AnnoyIndex
from scipy.sparse import csr_matrix, csc_matrix, hstack

from Cellist.Utility import *

def ImputationParser(subparsers):
    parser = subparsers.add_parser("impute", 
        help = "Perform spatially-aware gene imputation within each cluster. ")
    group_input = parser.add_argument_group("Input arguments")
    group_input.add_argument("--expr", dest = "expr", help = "Cell-level gene expression file in h5 format generated by 'Cellist seg'. ")
    group_input.add_argument("--spatial", dest = "spatial", help = "Spatial coordniates of cells generated by 'Cellist seg'. ")
    group_input.add_argument("--nworkers", dest = "num_workers", type = int, default = 8,
    help = "Maximum number of workers to use. ")

    group_output = parser.add_argument_group("Output arguments")
    group_output.add_argument("--outdir", dest = "out_dir",default = "",help = "Output directory.")
    group_output.add_argument("--outprefix", dest = "out_prefix",default = "",help = "Output prefix.")
    return(parser)

def ST_clustering(st_count_mat, st_count_genes, st_count_cells, spatial_df, out_dir, out_prefix):
    adata = anndata.AnnData(st_count_mat.transpose(), obs = dict(obs_names = st_count_cells), var = dict(var_names = st_count_genes))
    # preprocess
    sc.pp.filter_cells(adata, min_genes = 50)
    sc.pp.filter_genes(adata, min_cells = 100)
    sc.pp.calculate_qc_metrics(adata, inplace = True)
    # normalize data
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.highly_variable_genes(adata, flavor='seurat_v3', n_top_genes = 2000)
    sc.pp.scale(adata)
    sc.pp.pca(adata)
    sc.pp.neighbors(adata, n_pcs = 20, transformer=AnnoyTransformer(n_neighbors=15, n_trees=50))
    sc.tl.umap(adata)
    res = 0.6
    sc.tl.louvain(adata, resolution = res, key_added = 'louvain_%s' %res)
    # sc.tl.leiden(adata, n_iterations=2)
    fig, ax = plt.subplots(figsize = (6,6))
    sc.pl.umap(adata, color = 'louvain_%s' %res, ax = ax)
    plt.title('')
    plt.savefig(os.path.join(out_dir, "%s_louvain_UMAP.pdf" %out_prefix), bbox_inches='tight')
    # add celltype info
    adata.obsm["spatial"] = spatial_df.loc[adata.obs.index, ["y", "x"]].to_numpy()
    fig, ax = plt.subplots(figsize = (12,12))
    if adata.obs.shape[0] < 50000:
        scatter_size = 12
    else:
        scatter_size = 6
    sc.pl.embedding(adata, basis="spatial", color = 'louvain_%s' %res, s=scatter_size, show=False, ax = ax)
    ax.set_aspect('equal', adjustable='box')
    ax.invert_yaxis()
    plt.title('')
    plt.axis('off')
    plt.savefig(os.path.join(out_dir, "%s_louvain_spatial.pdf" %out_prefix), bbox_inches='tight')
    return(adata)

def CalAffinity_PCA(PCA_tree, i, k_params):
    nn, nd = PCA_tree.get_nns_by_item(i = i, n = k_params, include_distances=True)
    row_list = [i]*k_params
    col_list = nn
    dist_list = nd
    sigma = nd[int(k_params/3)]
    sigma_list = [sigma]*k_params
    return(np.array([row_list, col_list, dist_list, sigma_list]))

def CalAffinity_Spatial(Spatial_tree, i, k_params):
    nn, nd = Spatial_tree.get_nns_by_item(i = i, n = k_params, include_distances=True)
    row_list = [i]*k_params
    col_list = nn
    dist_list = nd
    sigma = nd[int(k_params/3)]
    sigma_list = [sigma]*k_params
    return(np.array([row_list, col_list, dist_list, sigma_list]))

def BuildTree(data, n_trees = 50):
    f = data.shape[1]
    t = AnnoyIndex(f, 'euclidean')
    n_samples = data.shape[0]
    for i in range(n_samples):
        v = data.iloc[i, :].tolist()
        t.add_item(i, v)
    t.build(n_trees)
    return(t)

def CalAffinityMat(result_list, n_samples):
    res_array = np.hstack(result_list)
    aff_array = np.exp(-np.square(res_array[2,:]/res_array[3,:]))
    aff_mat = csr_matrix((aff_array, (res_array[0,:], res_array[1,:])), shape=(n_samples, n_samples))
    aff_mat_sym = (aff_mat + aff_mat.T)/2
    return(aff_mat_sym)

def CalCombineAffinityMat(result_list, n_samples):
    res_array = np.hstack(result_list)
    aff_mat = csr_matrix((res_array[2,:], (res_array[0,:], res_array[1,:])), shape=(n_samples, n_samples))
    return(aff_mat)

def CombineAffinityMat(pca_aff_mat_sym, spatial_aff_mat_sym, knn, i):
    pca_nn = np.where(pca_aff_mat_sym[i,:].toarray()[0] > 0)[0]
    spatial_nn = np.where(spatial_aff_mat_sym[i,:].toarray()[0] > 0)[0]
    common_nn = np.intersect1d(pca_nn, spatial_nn)
    if common_nn.shape[0] > 5:
        if common_nn.shape[0] > knn:
            pca_aff_value = pca_aff_mat_sym[i, common_nn].toarray()[0]
            cut_off = np.quantile(pca_aff_value, (pca_aff_value.shape[0] - knn - 1)/pca_aff_value.shape[0])
            common_nn_cut = common_nn[pca_aff_value > cut_off]
        else:
            common_nn_cut = common_nn
        self_ind = np.where(common_nn_cut == i)[0]
        spatial_aff_array = spatial_aff_mat_sym[i, common_nn_cut].toarray()[0]
        spatial_aff_array[np.where(common_nn_cut != i)[0]] = spatial_aff_array[np.where(common_nn_cut != i)[0]]/np.sum(spatial_aff_array[np.where(common_nn_cut != i)[0]])
        spatial_aff_array[self_ind] = 1
        pca_aff_array = pca_aff_mat_sym[i, common_nn_cut].toarray()[0]
        pca_aff_array[np.where(common_nn_cut != i)[0]] = pca_aff_array[np.where(common_nn_cut != i)[0]]/np.sum(pca_aff_array[np.where(common_nn_cut != i)[0]])
        pca_aff_array[self_ind] = 1
        exp_weight = 0.5
        combined_aff_array = (1 - exp_weight)*spatial_aff_array + exp_weight*pca_aff_array
        col_arr = common_nn_cut
    else:
        pca_aff_value = pca_aff_mat_sym[i, pca_nn].toarray()[0]
        cut_off = np.quantile(pca_aff_value, (pca_aff_value.shape[0] - knn - 1)/pca_aff_value.shape[0])
        pca_nn_cut = pca_nn[pca_aff_value > cut_off]
        self_ind = np.where(pca_nn_cut == i)[0]
        pca_aff_array = pca_aff_mat_sym[i, pca_nn_cut].toarray()[0]
        pca_aff_array[np.where(pca_nn_cut != i)[0]] = pca_aff_array[np.where(pca_nn_cut != i)[0]]/np.sum(pca_aff_array[np.where(pca_nn_cut != i)[0]])
        pca_aff_array[self_ind] = 1
        combined_aff_array = pca_aff_array
        col_arr = pca_nn_cut
    row_arr = np.array([i]*combined_aff_array.shape[0])
    return(np.vstack([row_arr, col_arr, combined_aff_array]))

def read_h5(h5_file):
    expr_read = read_10X_h5(filename = h5_file)
    expr_mat = expr_read.matrix
    expr_genes = expr_read.names.tolist()
    expr_cells = expr_read.barcodes.tolist()
    if type(expr_genes[0]) == bytes:
        expr_genes = [i.decode() for i in expr_genes]
    if type(expr_cells[0]) == bytes:
        expr_cells = [i.decode() for i in expr_cells]
    return(expr_mat, expr_genes, expr_cells)

def Imputation(expr_file, spatial_file, num_workers, out_dir, out_prefix):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    expr_mat_all, expr_genes_all, expr_cells_all = read_h5(expr_file)
    spatial_df = pd.read_csv(spatial_file, sep = "\t", header = 0, index_col = 0)
    spatial_df.index = spatial_df.index.astype(str)
    adata = ST_clustering(expr_mat_all, expr_genes_all, expr_cells_all, spatial_df, out_dir, out_prefix + "_original")
    imputed_mat_list = []
    cells_list = []
    adata_meta_df = adata.obs
    data_pca = adata.obsm['X_pca'][:,:20]
    data_pca = pd.DataFrame(data_pca, index = adata.obs.index)
    cell_types = adata_meta_df.loc[:, 'louvain_0.6'].unique()
    for cell_type in cell_types:
        print(cell_type)
        adata_meta_df_ct = adata_meta_df.loc[adata_meta_df.loc[:,'louvain_0.6'] == cell_type,:]
        cells_ct = adata_meta_df_ct.index
        expr_mat_ct, expr_genes_ct, expr_cells_ct = sub_mat(expr_mat_all, expr_genes_all, expr_cells_all, cells_sub = cells_ct.tolist())
        data_pca_ct = data_pca.loc[expr_cells_ct,:]
        data_spatial_ct = spatial_df.loc[expr_cells_ct, ['x', 'y']]
        n_samples = data_pca_ct.shape[0]
        n_trees = 50
        k_params = 50
        PCA_tree = BuildTree(data = data_pca_ct, n_trees = n_trees)
        res_list = []
        with concurrent.futures.ThreadPoolExecutor(max_workers = num_workers) as executor:
            for i in range(n_samples):
                arg_tuple = (PCA_tree, i, k_params)
                res_list.append(executor.submit(CalAffinity_PCA, *arg_tuple))
            done, not_done = concurrent.futures.wait(res_list, timeout=None)
            PCA_result_list = [future.result() for future in done]
        pca_aff_mat_sym = CalAffinityMat(PCA_result_list, n_samples)
        k_params = 50
        Spatial_tree = BuildTree(data = data_spatial_ct, n_trees = n_trees)
        res_list = []
        with concurrent.futures.ThreadPoolExecutor(max_workers = num_workers) as executor:
            for i in range(n_samples):
                arg_tuple = (Spatial_tree, i, k_params)
                res_list.append(executor.submit(CalAffinity_Spatial, *arg_tuple))
            done, not_done = concurrent.futures.wait(res_list, timeout=None)
            Spatial_result_list = [future.result() for future in done]
        spatial_aff_mat_sym = CalAffinityMat(Spatial_result_list, n_samples)
        knn = 10
        res_list = []
        with concurrent.futures.ThreadPoolExecutor(max_workers = num_workers) as executor:
            for i in range(n_samples):
                arg_tuple = (pca_aff_mat_sym, spatial_aff_mat_sym, knn, i)
                res_list.append(executor.submit(CombineAffinityMat, *arg_tuple))
            done, not_done = concurrent.futures.wait(res_list, timeout=None)
            Combine_result_list = [future.result() for future in done]
        Combine_aff_mat = CalCombineAffinityMat(Combine_result_list, n_samples)
        Combine_aff_mat = csc_matrix(Combine_aff_mat)
        expr_mat_ct_imputed = Combine_aff_mat.dot(expr_mat_ct.T)
        print(expr_mat_ct_imputed.shape)
        imputed_mat_list.append(expr_mat_ct_imputed.T)
        cells_list = cells_list + expr_cells_ct
    imputed_mat_merged = hstack(imputed_mat_list)
    expr_mat_imputed_all, expr_genes_imputed_all, expr_cells_imputed_all = sub_mat(imputed_mat_merged, expr_genes_ct, cells_list, genes_sub = expr_genes_all, cells_sub = expr_cells_all)
    out_combined_file = os.path.join(out_dir, "%s_k_params_%s_knn_%s_imputed.h5" %(out_prefix, k_params, knn))
    out_h5ad_file = os.path.join(out_dir, "%s_k_params_%s_knn_%s_imputed_anndata.h5ad" %(out_prefix, k_params, knn))
    write_10X_h5(out_combined_file, expr_mat_imputed_all, features = expr_genes_imputed_all, barcodes = expr_cells_imputed_all, datatype = 'Gene')
    adata_imputed = ST_clustering(expr_mat_imputed_all, expr_genes_imputed_all, expr_cells_imputed_all, spatial_df, out_dir, out_prefix + "_imputed")
    adata_imputed.write_h5ad(out_h5ad_file)

if __name__ == '__main__':
    parser = ImputationParser()
    expr_file = parser.expr
    spatial_file = parser.spatial
    num_workers = parser.num_workers
    out_dir = parser.out_dir
    out_prefix = parser.out_prefix
    Imputation(expr_file, spatial_file, num_workers, out_dir, out_prefix)



# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2023-10-14 12:54:35
# @Last Modified by:   dongqing
# @Last Modified time: 2023-10-14 16:07:36


# -*- coding: utf-8 -*-
# @Author: Dongqing Sun
# @E-mail: Dongqingsun96@gmail.com
# @Date:   2021-06-10 15:25:08
# @Last Modified by:   Dongqing Sun
# @Last Modified time: 2021-08-03 21:36:20


import sys,os

try:
    from setuptools import setup, find_packages
except ImportError:
    print("Could not load setuptools. Please install the setuptools package.")

exec(open('src/Cellist/version.py').read())

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

def main():
    setup(
        name = "cellist",
        package_dir = {'':'src'},
        version = __version__,
        packages = find_packages(where="src"),
        scripts = ['bin/cellist'],
        package_data={
            "":["*.txt"]
        },
        install_requires = requirements,
        setup_requires = requirements,
        include_package_data = True,
        author = "Dongqing Sun",
        author_email = "Dongqingsun96@gmail.com",
        description = "Cellist (Cell identification in high-resolution Spatial Transcriptomics) is a cell segmentation tool for high-resolution spatial transcriptomics. ",
        license = "GPL-3.0",
        url = "https://github.com/dongqingsun96/Cellist",
        
        classifiers = [
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Natural Language :: English",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering :: Bio-Informatics"
        ],
        python_requires=">=3.8",
    )

if __name__ == "__main__":
    main()
